"""
Utility functions to transform YAML output.
"""
import sys
import csv
import itertools
from textwrap import wrap
from pprint import pprint

GENBANK, SRA, GFF = 1, 2, 3

usage = """

    reformat.py GENBANK|SRA|GFF
    
Sanity saving data re-formatter.
    
See https://www.biostarhandbook.com for further details.
    
"""
def extract_translation(argv, stream=sys.stdin):
    """
    Extracts the translation attribute of  a GFF file
    """
    if len(argv) < 4:
        sys.exit("**** extraction needs a type and a gene name")
    # Get the
    ftype = argv[2]
    target = argv[3]

    stream = itertools.takewhile(lambda x: not x.startswith(">"), stream)
    stream = filter(lambda x: not x.startswith('#'), stream)
    stream = csv.reader(stream, delimiter="\t")
    stream = filter(lambda x: x[2] == ftype, stream)

    for row in stream:
        elems = row[8].split(";")
        pairs = [elem.split("=") for elem in elems]
        attr = dict(pairs)
        name = attr.get("gene")
        if name == target:
            protein_id = attr.get("protein_id", "None")
            data_id = attr.get("ID", "None").split(".")[0]
            sequence = attr.get("translation", "")
            product = attr.get("product", "None")
            name = f">{protein_id} ID={data_id} gene={target} product={product}"
            print(name)

            text = "\n".join(wrap(sequence, width=80))
            print(text)


def print_genbank():
    """
    Prints GenBank rows from an NCBI YAML output.
    """
    data = parse_yaml()
    rows = data['genbank-sequences']

    # Fix up country and state
    for row in rows:

        # Simplify the needlessly obtuse and verbose column headers.
        row['country'] = row['locality'].get('country', '')
        row['state'] = row['locality'].get('state', '')
        row['region'] = row['gene-region']
        row['date'] = row['collection-date']

        # Replace the accession with refseq id if exists.
        row['accession'] = row.get('refseq-accession', row['accession'])

        # Remove of unused fields
        for key in ('locality', 'gene-region', 'refseq-accession', 'collection-date'):
            row.pop(key, None)

    # Print the Genbank table
    fieldnames = ['accession', 'region', 'date', 'country', 'state']
    writer = csv.DictWriter(sys.stdout, fieldnames=fieldnames, delimiter="\t")
    # writer.writeheader()
    writer.writerows(rows)


def print_sra():
    """
    Prints SRA rows from an NCBI YAML output.
    """
    # Fix up SRA
    data = parse_yaml()
    rows = data['sra-accessions']

    # Print the SRA run table.
    fieldnames = ['sra-run', 'bioproject', 'sra-experiment', 'sra-study', 'sra-sample', 'biosample']
    writer = csv.DictWriter(sys.stdout, fieldnames=fieldnames, delimiter="\t")
    # writer.writeheader()
    writer.writerows(rows)


def parse_yaml(stream=sys.stdin):
    """
    Parses a YAML stream.
    """
    try:
        # Don't make yaml a universal dependency.
        import yaml
    except ImportError as exc:
        print("*** This program requires pyyaml: pip install pyyaml", file=sys.stderr)
        sys.exit(1)

    # Parse the document
    data = yaml.load(stream, Loader=yaml.BaseLoader)

    # Uncomment to understand the structure of the data
    # pprint(data)

    if not data:
        sys.exit("*** error: empty stream")

    return data


def main():
    """
    Entry point for the script.
    """
    import argparse

    parser = argparse.ArgumentParser(description="Reformats bioinformatics data.", usage=usage)

    parser.add_argument("mode", type=str )

    args = parser.parse_args(sys.argv[1:])

    # Valid modes
    modes = dict(
        GENBANK=GENBANK,
        SRA=SRA,
        GFF=GFF,
    )

    # Select the mode
    mode = modes.get(args.mode.upper(), GENBANK)

    if mode == GENBANK:
        print_genbank()
    elif mode == SRA:
        print_sra()
    elif mode == GFF:
        extract_translation(sys.argv)


if __name__ == '__main__':
    main()





"""
Filters taxonomical information
"""
import csv
import re
import sys

usage = """

"""


def run_filter(blastinfo, lineage, text):

    # Take the regex from command line.
    patt = re.compile(text, re.IGNORECASE)

    # Open stream to lineage.
    stream1 = csv.reader(lineage, delimiter="\t")

    # Filter for more than a single column.
    stream1 = filter(lambda x: len(x) > 1, stream1)

    # Filter for second element matching the pattern.
    stream1 = filter(lambda x: patt.search(x[1]), stream1)

    # The taxids that match the words
    ids1 = map(lambda x: x[0], stream1)

    # Speeds up lookup.
    taxa = set(ids1)

    print (taxa)

    # Open stream to lineage.
    stream2 = csv.reader(blastinfo, delimiter=" ")

    # Filter for more than a single column.
    stream2 = filter(lambda x: len(x) > 2, stream2)

    # Filter for second element matching the pattern.
    stream2 = filter(lambda x: x[1] in taxa, stream2)

    for row in stream2:
        print (" ".join(row))


def main():
    """
    Entry point for the script.
    """
    import argparse

    parser = argparse.ArgumentParser(description="Filters taxonomical information.", usage=usage)

    parser.add_argument("blastinfo", type=argparse.FileType('r'))

    parser.add_argument("lineage", type=argparse.FileType('r'))

    parser.add_argument("pattern", type=str)

    args = parser.parse_args(sys.argv[1:])

    run_filter(blastinfo=args.blastinfo, lineage=args.lineage, text=args.pattern)


if __name__ == '__main__':
    main()



"""
A better version of the 'comm' command line tool.

Usage:

    comm.py A B

Selects columns from two files A and B and prints a three column output where:

1. the first column lists elements unique to A,
2. the second column lists elements unique to B,
3. the third column lists elements in both files.

Comments and empty lines are ignored.

Additional parameters may be used to

1. Select different columns from files.
2. Change the column delimiter.
3. Print only column 1, 2 or 3 of the output.

"""
import sys
import csv
import argparse
import itertools

FILL_VALUE = ''


def process(file1, file2, delimiter, colidx, args=None):
    """
    Processes the files and prints the output
    """

    def parse(stream):
        """
        A generator in a clojure to processes each stream.
        Returns the value of a column at the column index.
        """
        # Skip comment lines
        stream = filter(lambda x: not x.startswith('#'), stream)

        # Ignore empty lines.
        stream = filter(lambda x: x.strip(), stream)

        # Format the stream.
        stream = csv.reader(stream, delimiter=delimiter)

        # Generate empty values on missing columns.
        for row in stream:
            try:
                yield (row[colidx], None)
            except IndexError as exc:
                yield ('', None)

    # Make dictionaries to maintain original item order.
    store1 = dict(parse(file1))
    store2 = dict(parse(file2))

    # Generate the various groupings.
    union = [key for key in store1.keys() if key in store2]
    only1 = [key for key in store1.keys() if key not in store2]
    only2 = [key for key in store2.keys() if key not in store1]

    # Select output based on flags.
    if args.flag1:
        output = only1
    elif args.flag2:
        output = only2
    elif args.common:
        output = union
    else:
        output = itertools.zip_longest(only1, only2, union, fillvalue=FILL_VALUE)
        output = map(lambda x: "\t".join(x), output)

    # Print the output
    for line in output:
        print(line)


def main():
    """
    Entry point for the script.
    """
    parser = argparse.ArgumentParser(description='Finds common data in CSV/TAB delimited files')

    parser.add_argument('-c', action="store", dest="column", metavar="N", type=int,
                        help="column to process (%(default)s)", default=1)

    parser.add_argument('-t', action="store_true", dest="usetab", default=False,
                        help="tab delimited columns (default is comma)")

    parser.add_argument('-1', action="store_true", dest="flag1", default=False,
                        help="elements unique to file A (A-B)")
    parser.add_argument('-2', action="store_true", dest="flag2", default=False,
                        help="elements unique to file B (B-A)")
    parser.add_argument('-3', action="store_true", dest="common", default=False,
                        help="common elements in file (A and B)")

    parser.add_argument("file1", type=argparse.FileType('r'))
    parser.add_argument("file2", type=argparse.FileType('r'))

    if len(sys.argv) == 1:
        sys.argv.append("-h")

    args = parser.parse_args(sys.argv[1:])

    delimiter = "\t" if args.usetab else ","

    colidx = args.column - 1

    process(file1=args.file1, file2=args.file2, delimiter=delimiter, colidx=colidx, args=args)


if __name__ == '__main__':
    main()



"""
Iterates over a BAM file and prints reads that match a genotype at a location.

Works for matches and mistmatching bases only.
"""
import csv
import os
import sys

try:
    import pysam
except ImportError as exc:
    print("*** This program requires pysam: pip install pysam", file=sys.stderr)
    sys.exit(1)

def group_matches(fname, chrom, pos, metadata="metadata.txt"):
    if os.path.isfile(metadata):
        lines = csv.reader(open(metadata, 'rt'), delimiter="\t")
        pairs = [(p[0], p[2:]) for p in lines]
        store = dict(pairs)
    else:
        store = dict()

    # The position is zero based.
    pos = pos - 1
    bam = pysam.AlignmentFile(fname, 'rb')

    group = dict()
    for column in bam.pileup(chrom):
        if column.pos == pos:
            for read in column.pileups:
                name = read.alignment.query_name
                base = read.alignment.query_sequence[read.query_position]
                group.setdefault(base, []).append(name)

    for base, values in group.items():
        for name in values:
            name = name.split(".")[0]
            fields = store.get(name, [])
            meta = "\t".join(fields)
            data = [name, base, meta]
            print("\t".join(data))


def main():
    """
    Entry point for the script.
    """
    import argparse

    parser = argparse.ArgumentParser(description="Reformats various data")

    parser.add_argument("bam", type=str, )
    parser.add_argument("chrom", type=str, )
    parser.add_argument("pos", type=int, )

    args = parser.parse_args(sys.argv[1:])

    group_matches(fname=args.bam, chrom=args.chrom, pos=args.pos)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Combines Kallisto and Salmon abundance files.

Usage:

    cat SAMPLES | python combine.py WORKDIR 
    
Where:

- WORKDIR is a directory
- SAMPLES is a file with one SAMPLE_ID per line
 
For each SAMPLE_ID the program will attempt to read the file named located at:

    WORKDIR/SAMPLE_ID/abundances.tsv

Once loaded, the files will be combined into a tabular format printed to standard output. 
"""

import csv
import os
import sys

def parse(workdir, sample):
    """
    Returns a generator over the file.
    """

    # The path to the abundnace file.
    path = os.path.join(workdir, sample, "abundance.tsv")

    # Stop on missing files.
    if not os.path.isfile(path):
        print("ERROR: unable to read file path: %s" % path)
        sys.exit()

    # Open the stream.
    stream = csv.DictReader(open(path), delimiter="\t")

    # Generate the data
    for row in stream:
        yield row


def process(workdir, stream):
    """
    Processes and combines tsv files.
    """

    # The samples in the work directory.
    samples = [line.strip() for line in stream]

    collect = {}
    for sample in samples:

        # Open the stream to the file
        stream = parse(workdir, sample)

        for row in stream:

            # Identify data of interest.
            target_id, length, eff_length = row['target_id'], row['length'], row['eff_length']

            # Create a list for each new target id.
            if target_id not in collect:
                collect[target_id] = [target_id, length, eff_length]

            # Append the estimated counts to each target id.
            collect[target_id].append(row['est_counts'])

    # Print the new headers
    header = ["target_id", "length", "eff_length"] + samples
    header = "\t".join(header)
    print(header)

    # Go over the collection and print each row.
    for key, values in collect.items():
        line = "\t".join(values)
        print(line)


def main():
    """
    Entry point for the script.
    """
    import argparse

    parser = argparse.ArgumentParser(description="Combines Kallisto and Salmon abundance files")

    parser.add_argument("workdir", type=str, default=".", )

    args = parser.parse_args(sys.argv[1:])

    workdir = args.workdir

    if not os.path.isdir(workdir):
        print("ERROR! Path '%s' must be a directory" % workdir)
        sys.exit(-1)

    process(workdir=workdir, stream=sys.stdin)


if __name__ == '__main__':
    main()



import setuptools
import handbook

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="BiostarHandbook",
    version=handbook.version,
    author="Istvan Albert",
    author_email="istvan.albert@gmail.com",
    description="Utilities for the Biostar Handbook ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/biostars/biostar-handbook-code",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    scripts=[
        'scripts/foo.sh',
        'scripts/r/foo.r',
    ],
    entry_points={
        'console_scripts': [
            'comm.py=handbook.comm:main',
            'combine.py=handbook.combine:main',
            'reformat.py=handbook.reformat:main',
            'genotype.py=handbook.genotype:main',
            'search_taxa.py=handbook.search_taxa:main',
        ],
    },

    python_requires='>=3.6',

)
